<?php
if ( ! defined('DOKU_INC') ) define( 'DOKU_INC', realpath(dirname(__FILE__) . '/../../') . '/' ) ;
if ( ! defined('DOKU_PLUGIN') ) define( 'DOKU_PLUGIN', DOKU_INC . 'lib/plugins/' ) ;
require_once ( DOKU_PLUGIN . 'syntax.php' ) ;

/** Important Paragraf - Mark a paragraph with a class.

 Marks important paragraphs or sentences:

 An important paragraph starts with !!

 It is marked with &lt;div class="important" /&gt;

 A class name can be given in lower case ASCII between the !! and
 an optional |

 If you continue with a new paragraph that should still be
 important, end the line above with two spaces.

 It must be evaluated before the dl plugin.

 License: GPL
 */
class syntax_plugin_important_paragraf extends DokuWiki_Syntax_Plugin {

    function getInfo() {
        return array( 'author' => 'Marc Wckerlin', 'email' => 'marc [at] waeckerlin [dot-org]', 'name' => 'Important Paragraf', 'desc' => 'mark paragrafs as !!importtant', 'url' => 'http://marc.waeckerlin.org' ) ;
    }

    function getType() {
        return 'container' ;
    }

    function getSort() {
        return 2 ;
    }

    function accepts( $mode ) {
        if ( ! count($this->allowedModes) ) {
            global $PARSER_MODES ;
            $this->allowedModes = array_merge( $PARSER_MODES['container'], $PARSER_MODES['baseonly'], $PARSER_MODES['formatting'], $PARSER_MODES['substition'], $PARSER_MODES['protected'], $PARSER_MODES['disabled'], $PARSER_MODES['paragraphs'] ) ;
            unset( $this->allowedModes[array_search('preformatted', $this->allowedModes)] ) ;
        }
        return parent::accepts( $mode ) ;
    }

    function connectTo( $mode ) {
        $this->Lexer->addEntryPattern( '^ *\!\![a-z]+\|', $mode, 'plugin_important_paragraf' ) ;
        $this->Lexer->addEntryPattern( '^ *\!\!', $mode, 'plugin_important_paragraf' ) ;
    }

    function postConnect() {
        $this->Lexer->addPattern( ' (?= \n)', 'plugin_important_paragraf' ) ;
        $this->Lexer->addExitPattern( '(?<! )\n', 'plugin_important_paragraf' ) ;
    }

    function handle( $match, $state, $pos, &$handler ) {
        return array( $match, $state ) ;
    }

    function render( $format, &$renderer, $data ) {
        list( $match, $state ) = $data ;
        switch ( $state ) {
            case DOKU_LEXER_ENTER:
                {
                    if ( strlen($match) > 3 ) $class = substr( $match, 2, strlen($match) - 3 ) ;
                    else  $class = 'important' ;
                    $renderer->doc .= '<div class="' . $class . '"><div>' ;
                }
                return true ;
            case DOKU_LEXER_EXIT:
                {
                    $renderer->doc .= '</div></div>' ;
                }
                return true ;
            case DOKU_LEXER_MATCHED:
                {
                    $renderer->doc .= '</div><div>' ;
                    return true ;
                }
            case DOKU_LEXER_UNMATCHED:
                {
                    $renderer->doc .= htmlspecialchars( $match ) ;
                }
                return true ;
        }
        return false ;
    }
}
?>